# -*-coding:utf-8-*-

import collections
import numpy as np
import itertools
from . import globalVariable as g


get_bigger = lambda a, b: a if a > b else b


class QuantumCircuit:
	def __init__ (self, **kwargs):
		# device size 
		self.device_size = kwargs.get("layout_size")
		if self.device_size is None:
			raise error.Error("self.de is not provided.")

		# pre-determined qubit table
		if "qubit_table" in kwargs: 
			qubit_table = kwargs.get("qubit_table")
		
		# synthesis option
		self.synthesis_option = kwargs.get("option")
		
		# flag for allowing swap gate or not
		self.flag_swap = self.synthesis_option.get("allow_swap")
		if self.flag_swap is None:
			self.flag_swap = False

		# flag for allowing parallel swap
		self.parallel_swap = self.synthesis_option.get("parallel_swap")
		if self.parallel_swap is None:
			self.parallel_swap = False

		
		self.inverse_mapping_table = {v: k for k, v in qubit_table.items()}
		
		# system code
		self.system_code = {"cbit": [], "circuit": [], "initial_mapping": qubit_table}
		
		
		# list of algorithm qubits
		self.flag_algorithm_qubits = False
		algorithm_qubits = kwargs.get("algorithm_qubits")
		if algorithm_qubits is not None:
			self.system_code["qubit"] = algorithm_qubits
			self.flag_algorithm_qubits = True
		
		self.list_algorithm_qubits = set([])
		
		# variables for function list and circuit depth
		self.function_list = collections.defaultdict(int)
		self.operation_index = collections.defaultdict(int)


	def get_qubit_association(self):
		return self.qubit_association


	def one_qubit_gate(self, gate, qubit, **kwargs):
		"""
			function to process 1-qubit gate
		"""
		# add the gate to the function list : increment by 1
		self.function_list[gate] += 1

		# add the qubit to the list of algorithm qubit
		if not self.flag_algorithm_qubits:
			self.list_algorithm_qubits.add(qubit)

		# form a system code and append it to the system code list
		if gate in [g.str_gate_rz, g.str_gate_rx, g.str_gate_ry]:
			# rotational gate in one axis : checking angle
			if "angle" in kwargs: angle = kwargs["angle"]
			_str_code = [gate, str(angle), str(qubit)]

		elif gate in [g.str_gate_phase]:
			# special rotational gate 
			if "angle" in kwargs: angle = kwargs["angle"]
			_str_code = [gate, str(angle), str(qubit)]

		elif gate in [g.str_gate_u]:
			# rotational gate in three axes : checking angle x, z, y
			if "angle_x" not in kwargs or "angle_y" not in kwargs or "angle_z" not in kwargs:
				raise error.Error("Error Happened.")
			else:
				_str_code = [gate, kwargs["angle_x"], kwargs["angle_y"], kwargs["angle_z"], qubit]

		elif gate == g.str_gate_measz:
			cbit = kwargs.get("cbit") 
			if cbit is not None:
				self.system_code["cbit"].append(cbit)
				_str_code = [gate, str(qubit), str(cbit)]
			else:
				_str_code = [gate, str(qubit), str(qubit)]
		
		else:
			_str_code = [gate, str(qubit)]
		

		self.system_code["circuit"].append(_str_code)
		self.operation_index[qubit]+=1
		

	def swap(self, path, **kwargs):
		"""
			function process 2-qubit gate SWAP
		"""
		qubit1, qubit2, weight = path[:]
		
		if qubit1 == qubit2:
			raise error.Error("For SWAP, qubit1 and qubit2 are the same. ")

		# if a swap gate is applicable	
		if self.flag_swap:
			self.function_list[g.str_gate_swap]+=1

			physical_apply_index = get_bigger(self.operation_index[qubit1], self.operation_index[qubit2]) + 1
			for qubit in [qubit1, qubit2]:
				self.operation_index[qubit] = physical_apply_index

			str_code = [g.str_gate_swap, str(qubit1), str(qubit2)]
			self.system_code["circuit"].append(str_code)

		# otherwise, it should be substituted by 3 cnot gates
		else:
			if weight == 1:
				# ctrl -> trgt
				self.cnot([qubit1, qubit2])
				
				self.one_qubit_gate(g.str_gate_h, qubit1)
				self.one_qubit_gate(g.str_gate_h, qubit2)
				
				self.cnot([qubit1, qubit2])

				self.one_qubit_gate(g.str_gate_h, qubit1)
				self.one_qubit_gate(g.str_gate_h, qubit2)
				
				self.cnot([qubit1, qubit2])

			elif weight == 3:
				# ctrl <- trgt
				self.cnot([qubit2, qubit1])

				self.one_qubit_gate(g.str_gate_h, qubit1)
				self.one_qubit_gate(g.str_gate_h, qubit2)

				self.cnot([qubit2, qubit1])

				self.one_qubit_gate(g.str_gate_h, qubit1)
				self.one_qubit_gate(g.str_gate_h, qubit2)

				self.cnot([qubit2, qubit1])

			elif weight == 4:
				self.cnot([qubit1, qubit2])
				self.cnot([qubit2, qubit1])
				self.cnot([qubit1, qubit2])

		qubit_table = kwargs.get("qubit_table")
		if qubit_table is not None:
			inverse_mapping_table = {v: k for k, v in qubit_table.items()}

			inverse_mapping_table[qubit1], inverse_mapping_table[qubit2] =\
				inverse_mapping_table[qubit2], inverse_mapping_table[qubit1]

			return {v: k for k, v in inverse_mapping_table.items()}


	def manage_cnot(self, qubits, **kwargs):
		"""
			function to process a CNOT gate
			if necessary, manage to run a sequene of SWAP gates
		"""
		list_path = kwargs.get("path")
		qubit_mapping_table = kwargs.get("qubit_table")
		final_operation = kwargs.get("final_operation")

		if list_path is None:
			raise error.Error("For CNOT, qubit move path is not provided.")
		if qubit_mapping_table is None:
			raise error.Error("For CNOT, qubit mapping table is not provided.")
		if final_operation is None:
			raise error.Error("Final operation is not provided.")

		if not self.flag_algorithm_qubits:
			self.list_algorithm_qubits.add(qubits[0])
			self.list_algorithm_qubits.add(qubits[1])

		# cnot path 의 사이즈가 1보다 크뎐, swap 이 필요함
		if len(list_path)>1:
			# cnot path 상에서 병렬적 swap 이 가능한 경우
			# 병렬적 swap : forward path: cnot path 의 처음부터 순서대로
			#			: backward path: cnot path 의 마지막 부터 역순으로
			if self.parallel_swap:
				# forward path, backward path 그리고 마지막 cnot 을 구하기 위해
				# 먼저 cnot path의 길이를 확인하고, 그 절반 값을 기준으로 삼는다.
				length_swap_path = len(list_path)
				half_length = int(length_swap_path/2)

				# cnot path의 처음부터 절반까지는 forward swap path
				forward_path = list_path[:half_length]
				# cnot path의 마지막 부터 절반 까지는 backward swap path
				backward_path = reversed(list_path[half_length+1:])
				# cnot path의 가운데 있는 큐빗 pair 는 최종 cnot 수행할 큐빗 인덱스
				final_cnot = list_path[half_length]

				# forward path, backward path 를 각각 순서대로 swap 수행함
				for item in itertools.zip_longest(forward_path, backward_path):
					if item[0] is not None: 
						qubit_mapping_table = self.swap(item[0], qubit_table=qubit_mapping_table)
					if item[1] is not None:
						qubit_mapping_table = self.swap(item[1], qubit_table=qubit_mapping_table)
			
			# 병렬적 swap이 허용되지 않으면, cnot path 상에서 순서대로 serial swap 수행함
			# cnot path 상의 마지막 item 은 최종 cnot 수행할 큐빗 pair
			else:
				swap_path = list_path[:-1]
				final_cnot = list_path[-1]

				# swap path 상에서 순서대로 swap 수행함
				for edge in swap_path:
					qubit_mapping_table = self.swap(edge, qubit_table=qubit_mapping_table)

				self.inverse_mapping_table = {v: k for k, v in qubit_mapping_table.items()}

		else:
			# cnot path 테이블의 길이가 1이면, 바로 cnot 수행하면 된다.
			final_cnot = list_path[0]
		
		# 최종 cnot 수행
		if final_operation == "CNOT":  self.cnot(final_cnot[:2])

		# 최종 swap 수행
		elif final_operation == "SWAP": 
			qubit_mapping_table = self.swap(final_cnot, qubit_table=qubit_mapping_table)
		
		# 원거리상의 2-큐빗 게이트 수행 종료 후의 큐빗 매핑 테이블
		self.inverse_mapping_table = {v: k for k, v in qubit_mapping_table.items()}
		return qubit_mapping_table


	def cnot(self, qubits):
		"""
			a function to process a CNOT gate
			args : list of two qubits, ctrl and trgt
		"""
		ctrl, trgt = map(int, qubits)

		# if ctrl qubit and trgt qubit are identical, raise error 
		if ctrl == trgt:
			raise error.Error("For CNOT, ctrl qubit and trgt qubit are the same. {}".format(qubits))
			
		
		self.function_list[g.str_gate_cnot] += 1

		physical_apply_index = get_bigger(self.operation_index[ctrl], self.operation_index[trgt]) + 1

		# update data
		for qubit in [ctrl, trgt]:
			self.operation_index[qubit] = physical_apply_index


		# synthesis mode 가 system code 생성이면
		str_code = [g.str_gate_cnot, str(ctrl), str(trgt)]
		self.system_code["circuit"].append(str_code)

	

	def get_analysis(self):
		"""
			function to return an analyzed static data about a circuit
		"""
		# circuit depth
		physical_circuit_depth = max(list(self.operation_index.values()))
		
		# list of qubits used in algorithm
		if not self.flag_algorithm_qubits:
			self.system_code["qubit"] = list(self.list_algorithm_qubits)

		# number of qubits really used in the quantum chip
		used_qubits = len([k for k, v in self.operation_index.items() if v > 0])

		# KQ is evaluated from the really used qubits
		return {"Qubits": {"Algorithm": len(self.system_code["qubit"]), 
						   "Device": self.device_size, 
						   "Used": used_qubits},
			    "Circuit Depth" : physical_circuit_depth,
			    "KQ": int(used_qubits * physical_circuit_depth),
			    "Function List": self.function_list}


	def get_system_code(self, **kwargs):
		"""
			function to get a system code
			if file name is provided then, the system code is exported to that file too.
		"""

		path_syscode = kwargs.get("file")
		if path_syscode is not None:
			with open(path_syscode, "w") as outfile:
				for i in self.system_code:
					outfile.write("{}\n".format(i))

		# qubits and cbits for the algorithm (circuit)
		self.system_code["cbit"] = list(set(self.system_code["cbit"]))
		self.system_code["initial_mapping"] = {k: v for k, v in self.system_code["initial_mapping"].items() 
														if k in self.system_code["qubit"]}
		
		return {"system_code": self.system_code}