
import re
import simplejson as json
import math
from . import circuit
from . import graph
from . import error
from . import qubitmapping
from . import formatconversion
from . import globalVariable as g

parser = re.compile("[\{\[\]a-zA-Z0-9_.*\->\+}]+")


def synthesize(qasm_code, **kwargs):
	# over several loops
	# qubit table 생성 -> random
	# synthesize_flattened_qasm 함수 호출... 최소 time... target function

	# inputs for the circuit synthesis
	arguments = kwargs.get("arguments")
	if arguments is None:
		raise error.Error("arguments is not provided")

	# quantum chip information
	file_qchip = arguments.get("qchip")
	if file_qchip is None:
		raise error.Error("qchip information is not provided")
	else:
		json_qchip_data = open(file_qchip).read()
		qchip_data = json.loads(json_qchip_data)
		qchip_size = len(qchip_data["qubit_connectivity"])

	# circuit synthesis option
	synthesis_option = arguments.get("option")
	if synthesis_option is None: 
		raise error.Error("synthesis option is not provided.")

	# given qubit mapping table
	file_qubit_mapping = arguments.get("qubit_table")
	if file_qubit_mapping is not None:
		flag_initial_qubit_table = True
		json_data = open(file_qubit_mapping).read()
		qubit_mapping = json.loads(json_data)
	else:
		flag_initial_qubit_table = False

	
	# list of qubits used in quantum algorithm
	list_algorithm_qubits = []

	# the number of cnot gates in algorithm
	cnot_counts_in_algorithm = 0
	
	# pre-analyze the numbers of algorithms qubits and cnot gates
	with open(qasm_code, "r") as infile:
		for line in infile:
			token = parser.findall(line)
			if not len(token): continue
			if token[0] in ["Qubit", "qubit"]: list_algorithm_qubits.append(token[1])
			elif token[0] in [g.str_gate_cnot]: cnot_counts_in_algorithm+=1
			elif token[0] in [g.str_gate_swap]: cnot_counts_in_algorithm+=3
	
	if not len(list_algorithm_qubits):
		with open(qasm_code, "r") as infile:
			for line in infile:
				token = parser.findall(line)
				if not len(token): continue
				if token[0] in [g.str_gate_cnot, g.str_gate_swap, g.str_gate_cz]:
					list_algorithm_qubits.extend(token[1:3])
				elif token[0] in [g.str_gate_rz]:
					list_algorithm_qubits.append(token[2])
				else:
					list_algorithm_qubits.append(token[1])

		list_algorithm_qubits = list(set(list_algorithm_qubits))

	# pre-set best performance data
	best_performance = math.inf
	best_performance_data = None

	# setting the number of mapping iterations
	iteration = synthesis_option.get("iteration")
	if iteration is None: iteration = 10
	else:
		iteration = int(iteration)

	# qubit mapping table 이 주어지면, 한 번만 매핑하면 됨
	# 만약, qubit mapping table 이 주어지지 않으면, 랜덤으로 생성해서 반복적으로 매핑함
	if flag_initial_qubit_table:
		synthesized_result = synthesize_flattened_qasm(qasm_code, 
														qchip=file_qchip,
											 	   		qubit_table=qubit_mapping,
											 	   		option=synthesis_option)
		
		if float(synthesized_result["analysis"]["Circuit Depth"]) < best_performance:
			best_performance = float(synthesized_result["analysis"]["Circuit Depth"])
			best_performance_data = synthesized_result

	else:
		# perform the mapping as much as the number of **iterations**
		for iter_idx in range(iteration):
			# random seed for generating random numbers
			random_seed = synthesis_option.get("random_seed")

			if random_seed is None:
				qubit_mapping = qubitmapping.initialize_qubit_mapping(list_algorithm_qubits, qchip_size, 
																		option=synthesis_option["initial_mapping_option"])
			else:	
				qubit_mapping = qubitmapping.initialize_qubit_mapping(list_algorithm_qubits, qchip_size, 
																		option=synthesis_option.get("initial_mapping_option"),
																		seed=int(random_seed))
			
			synthesized_result = synthesize_flattened_qasm(qasm_code, 
															qchip=file_qchip,
												 	   		qubit_table=qubit_mapping,
												 	   		option=synthesis_option,
												 	   		algorithm_qubits=list_algorithm_qubits)
			

			if float(synthesized_result["analysis"]["Circuit Depth"]) < best_performance:
				best_performance = float(synthesized_result["analysis"]["Circuit Depth"])
				best_performance_data = synthesized_result


	# post processing for the analysis
	# 1. number of cnot : the effectiveness of the circuit synthesis
	cnot_counts_in_syscode = best_performance_data["analysis"]["Function List"]["CNOT"] +\
					3*best_performance_data["analysis"]["Function List"]["SWAP"]

	best_performance_data["analysis"]["CNOT Overhead"] = {"Algorithm": cnot_counts_in_algorithm,
														  "Circuit": cnot_counts_in_syscode,
														  "Overhead": cnot_counts_in_syscode - cnot_counts_in_algorithm}

	##################################################################
	# 합성 목적에 따라 후속 작업 수행
	# - 시스템 코드: 1. time ordered format 으로 변환
	#             2. analysis data 타입 포맷 변환 (for dumping the data into json file)
	##################################################################
	# transform the simple list --> time ordered system code
	best_performance_data["system_code"]["circuit"] =\
		formatconversion.transform_ordered_syscode(best_performance_data["system_code"]["circuit"], qchip_size)


	return best_performance_data


def synthesize_flattened_qasm(qasm_code, **kwargs):

	# extract the quantum chip information (file)
	file_qchip = kwargs.get("qchip")
	if file_qchip is None:
		raise error.Error("Qchip information is not given.")
	
	# extract qubit mapping table that is provided as input
	qubit_mapping_table = kwargs.get("qubit_table")
	if qubit_mapping_table is None:
		raise error.Error("Qubit Mapping Table is not provided.")

	# circuit synthesis option
	synthesis_option = kwargs.get("option")
	if synthesis_option is None:
		raise error.Error("Synthesis Option is not provided.")

	# list of qubits used in algorithm (extracted from QASM)
	list_algorithm_qubits = kwargs.get("algorithm_qubits")
	if list_algorithm_qubits is not None:
		flag_algorithm_qubits = True
	else:
		list_algorithm_qubits = set([])

	if isinstance(qubit_mapping_table, str):
		json_data = open(qubit_mapping_table).read()
		qubit_mapping_table = json.loads(json_data)
	
	# take quantum chip data
	json_data = open(file_qchip).read()
	qchip_data = json.loads(json_data)
	qubit_connectivity = {int(k): v for k, v in qchip_data["qubit_connectivity"].items()}
	
	# initialize quantum circuit object that is responsible to gather static information about the circuit
	_circuit = circuit.QuantumCircuit(layout_size=len(qubit_connectivity.keys()), 
									  qubit_table=qubit_mapping_table,
									  option=synthesis_option,
									  algorithm_qubits=list_algorithm_qubits)

	g.initialize_globals()

	# over all the quantum assembly code instruction
	with open(qasm_code, "r") as infile:
		for line in infile:
			# read and parse the string line (instruction)
			tokens = parser.findall(line)
			if not len(tokens): continue
			gate = tokens[0]

			# case 1: one-qubit gate (rotational gate with angle, measz(measx) with classical bit, the other without other args)
			if gate in g.list_one_qubit_gates:
				if gate in [g.str_gate_rz, g.str_gate_rx, g.str_gate_ry]:
					qubit, angle = tokens[1:]
					try:
						if type(eval(qubit)) == float: angle, qubit = qubit, angle
					except: pass

					qubit_idx = int(qubit_mapping_table[qubit])
					_circuit.one_qubit_gate(gate, qubit_idx, angle=float(angle))

					if not flag_algorithm_qubits:
						list_algorithm_qubits.add(qubit)

				elif gate in [g.str_gate_measz, g.str_gate_measx]:
					if len(tokens) == 4:
						qubit, cbit = tokens[1::2]
						qubit_idx = int(qubit_mapping_table[qubit])
						_circuit.one_qubit_gate(gate, qubit_idx, cbit=cbit)

					elif len(tokens) == 2:
						qubit = tokens[1]
						qubit_idx = int(qubit_mapping_table[qubit])
						_circuit.one_qubit_gate(gate, qubit_idx)
					
					if not flag_algorithm_qubits:
						list_algorithm_qubits.add(qubit)

				else:
					qubit = tokens[1]
					qubit_idx = int(qubit_mapping_table[qubit])
					_circuit.one_qubit_gate(gate, qubit_idx)

					if not flag_algorithm_qubits:
						list_algorithm_qubits.add(qubit)

			# case 2: two-qubit gate: cnot 과 swap 고려함
			elif gate in g.list_two_qubit_gates:
				if gate in [g.str_gate_cnot]:
					ctrl, trgt = tokens[1:]
					ctrl_idx = int(qubit_mapping_table[ctrl])
					trgt_idx = int(qubit_mapping_table[trgt])

					# find the qubit move path by exploiting dijkstra's shortest path search algorithm
					connection_path = graph.findQubitConnectionPath(qubit_connectivity, ctrl_idx, trgt_idx)
					
					# final_cnot: 원거리 이동후 마지막에 cnot 수행 여부 결정
					qubit_mapping_table = _circuit.manage_cnot([ctrl_idx, trgt_idx], path=connection_path, 
																					 qubit_table=qubit_mapping_table,
					
																					 final_operation="CNOT")
					if not flag_algorithm_qubits:
						list_algorithm_qubits.add(ctrl)
						list_algorithm_qubits.add(trgt)

				elif gate in [g.str_gate_swap]:
					ctrl, trgt = tokens[1:]
					ctrl_idx = int(qubit_mapping_table[ctrl])
					trgt_idx = int(qubit_mapping_table[trgt])

					qubit_mapping_table = _circuit.swap([ctrl_idx, trgt_idx, 4])

					if not flag_algorithm_qubits:
						list_algorithm_qubits.add(ctrl)
						list_algorithm_qubits.add(trgt)
					

	qubit_mapping_table = {k: v for k, v in qubit_mapping_table.items() if k in list_algorithm_qubits}

	system_code = _circuit.get_system_code()["system_code"]
	system_code["final_mapping"] = qubit_mapping_table

	return {"system_code": system_code,
		    "qchip": qchip_data,
		    "analysis": _circuit.get_analysis()}

if __name__ == "__main__":
	import os
	from pprint import pprint

	path_qasm = os.path.join("../tests/resources/DB-Assembly", "3_17_14.qasmf")
	path_qchip = os.path.join("../tests/resources/DB-QChip", "file_qchip_17.json")
	
	synthesis_option = {'option': {'allow_swap': True,
						            'goal': 'system_code',
						            'iteration': 10,
						            'qubit_connectivity': 'user',
						            "initial_mapping_option": "random"},

	 'path_job': 'DB-Jobs/Synthesis-hwang-yongsus-MacBook-Pro.local-20210525182020',
	 'qchip': path_qchip}

	ret = synthesize(qasm_code=path_qasm, arguments=synthesis_option)
	pprint(ret)
